﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ASTRO
{
    public class SwissEph
    {
       public Extensions ext;
     
  
//        #ifdef TRANSITS
//#ifdef TEST_ITERATIONS
  /**
  * Returns the number of iterations of the last transit calculation. The
  * transit calculations calculate a planet's position and proceed calculating
  * the planets positions until the maximum precision has been reached. This
  * method returns the count of calculations performed.<p>
  * This method requires precompilation of the original sources with the
  * -DTEST_ITERATIONS switch.<p>
  * ATTENTION: This method is mainly for debugging and testing purposes, and
  * is in NO WAY thread save and the count of iterations is ONLY valid for
  * the last transit calculation before any new calculation has started!
  * @return Number of iterations for the last transit calculation.
#ifdef EXTPRECISION
  * @see swisseph.SwissEph#getTransitET(int, double, int, double, double, boolean)
  * @see swisseph.SwissEph#getTransitUT(int, double, int, double, double, boolean)
#else
  * @see swisseph.SwissEph#getTransitET(int, double, int, double, boolean)
  * @see swisseph.SwissEph#getTransitUT(int, double, int, double, boolean)
#endif /* EXTPRECISION */
  //*/
  public long getIterateCount() {
    if (ext==null) 
    {
        ext=new Extensions(this); 
    }
    return ext.getIterateCount();
  }


  /**
  * Searches for the next or previous transit of a planet over a specified
  * longitude, latitude, distance or speed value with geocentric or topocentric
  * positions in a tropical or sidereal zodiac. Dates are interpreted as ET
  * (=UT&nbsp;+&nbsp;deltaT).<p>
  * See swisseph.TCPlanet or swisseph.TCPlanetPlanet below for examples on
  * how to use this method.<p>
  *
  * @param tc The TransitCalculator that should be used here.
  * @param jdET The date (and time) in ET, from where to start searching.
  * @param backwards If backward search should be performed.
  * @return return A double containing the julian day number for the next /
  * previous transit as ET.
  * @see swisseph.TCPlanet
  * @see swisseph.TCPlanetPlanet
  */
  public double getTransitET(TransitCalculator tc, double jdET, bool backwards)
         /*throws IllegalArgumentException, SwissephException*/ {
    return getTransitET(tc,
                        jdET,
                        backwards,
                        (backwards?-Double.MaxValue:Double.MaxValue));
  }
  /**
  * Searches for the next or previous transit of a planet over a specified
  * longitude, latitude, distance or speed value with geocentric or topocentric
  * positions in a tropical or sidereal zodiac. Dates are interpreted as ET
  * (=UT&nbsp;+&nbsp;deltaT).<p>
  * See swisseph.TCPlanet or swisseph.TCPlanetPlanet below for examples on
  * how to use this method.<p>
  *
  * @param tc The TransitCalculator that should be used here.
  * @param jdET The date (and time) in ET, from where to start searching.
  * @param backwards If backward search should be performed.
  * @param jdLimit This is the date, when the search for transits should be
  * stopped, even if no transit point had been found up to then.
  * @return return A double containing the julian day number for the next /
  * previous transit as ET.
  * @see swisseph.TCPlanet
  * @see swisseph.TCPlanetPlanet
  */

  public double getTransitET(TransitCalculator tc, double jdET, bool backwards, double jdLimit)
         /*throws IllegalArgumentException, SwissephException*/ {
    if (ext==null) { ext=new Extensions(this); }
    bool calcUT = (tc is TCHouses);
    return ext.getTransit(tc, jdET - (calcUT ? Swe.Function.swe_deltat(jdET) : 0), backwards, jdLimit) +
            (calcUT ? Swe.Function.swe_deltat(jdET) : 0);
  }
  /**
  * Searches for the next or previous transit of a planet over a specified
  * longitude, latitude, distance or speed value with geocentric or topocentric
  * positions in a tropical or sidereal zodiac. Dates are interpreted as UT
  * (=ET&nbsp;-&nbsp;deltaT).<p>
  * See swisseph.TCPlanet or swisseph.TCPlanetPlanet below for examples on
  * how to use this method.<p>
  *
  * @param tc The TransitCalculator that should be used here.
  * @param jdUT The date (and time) in UT, from where to start searching.
  * @param backwards If backward search should be performed.
  * @return return A double containing the julian day number for the next /
  * previous transit as UT.
  * @see swisseph.TCPlanet
  * @see swisseph.TCPlanetPlanet
  */
  
  public double getTransitUT(
          TransitCalculator tc,
          double jdUT,
          bool backwards)
         /*throws IllegalArgumentException, SwissephException */ {
    if (ext==null) { ext=new Extensions(this); }
    bool calcUT = (tc is TCHouses);

    
      
    double jdET = ext.getTransit(
                          tc,
                          jdUT + (calcUT ? 0 : Swe.Function.swe_deltat(jdUT)),
                          backwards,
                          (backwards?-Double.MaxValue:Double.MaxValue));


    return jdET - (calcUT ? 0 : Swe.Function.swe_deltat(jdET));
  }
  /**
  * Searches for the next or previous transit of a planet over a specified
  * longitude, latitude, distance or speed value with geocentric or topocentric
  * positions in a tropical or sidereal zodiac. Dates are interpreted as UT
  * (=ET&nbsp;-&nbsp;deltaT).<p>
  * See swisseph.TCPlanet or swisseph.TCPlanetPlanet below for examples on
  * how to use this method.<p>
  *
  * @param tc The TransitCalculator that should be used here.
  * @param jdUT The date (and time) in UT, from where to start searching.
  * @param backwards If backward search should be performed.
  * @param jdLimit This is the date, when the search for transits should be
  * stopped, even if no transit point had been found up to then. It is
  * interpreted as UT time as well.
  * @return return A double containing the julian day number for the next /
  * previous transit as UT.
  * @see swisseph.TCPlanet
  * @see swisseph.TCPlanetPlanet
  */
  public double getTransitUT( TransitCalculator tc,   double jdUT, bool backwards,  double jdLimit)
         /*throws IllegalArgumentException, SwissephException*/ 
  {
    if (ext==null)
    {
        ext=new Extensions(this);
    }

    double jdET = ext.getTransit(tc,jdUT + Swe.Function.swe_deltat(jdUT),  backwards, jdLimit + Swe.Function.swe_deltat(jdLimit));
    
    return jdET - Swe.Function.swe_deltat(jdET);
  }
//#endif /* TRANSITS */
    }
}
