﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ASTRO
{
    public  class TCPlanetPlanet : TransitCalculator
    {
        /**
* This class implements a TransitCalculator for two planets
* in relative positions or speeds to each other.<p>
* You would create a TransitCalculator from this class and
* use the SwissEph.getTransit() methods to actually calculate
* a transit, e.g.:<p>
* <pre>
* SwissEph sw = new SwissEph(...);
* ...
* int flags = Swe.Constant.SEFLG_SWIEPH |
*             Swe.Constant.SEFLG_TRANSIT_LONGITUDE;
* bool backwards = true;
* 
* TransitCalculator tc = new TCPlanetPlanet(
*                                  sw,
*                                  Swe.Constant.SE_MERCURY,
*                                  Swe.Constant.SE_VENUS,
*                                  flags,
*                                  0);
* ...
* double nextTransitUT = sw.getTransitUT(tc, jdUT, backwards);
* </pre>
* This would calculate the last (UT-) date, when Mercury and Venus
* had the same longitudinal position.
*/
  private int pl1, pl2;
  private int idx = 0; // The index into the xx[] array in swe_calc() to use:
  private int tflags = 0; // The transit flags
  private int flags = 0;  // The calculation flags for swe_calc()
  private bool calcPartile = false;
  private bool calcNonPartile = false;
  private bool calcPartileBoth = false;
  private bool isPartile = false;
  private bool calcYoga = false;
  private double minSpeed1, maxSpeed1;
  private double minSpeed2, maxSpeed2;
  private double minSpeed, maxSpeed;
  // The y = f(x) value to reach, speaking mathematically...
  private double offset = 0.0;
//#ifdef EXTPRECISION
  private double extPrecision = 1.0;
//#endif /* EXTPRECISION */

  private double lon1=0, lon2=-1000;	// For partile aspects only


  double minVal = 0.0, maxVal = 0.0;  // Thinking about it...


  /**
  * Creates a new TransitCalculator for relative transits of two different
  * planets to each other with the option for transits over longitudes,
  * latitudes, distance or the speed in any of these directions in the
  * geocentric or topocentric coordinate system, and in tropical or sidereal
  * zodiac system, both with the sum and difference of both planets positions
  * and speeds.<p>
  * Calculation of partile transits is possible now, but it's still a
  * rather slow calculation and not finally tested.<p>
  * @param sw A SwissEph object, if you have one available. Can be null.
  * @param pl1 The first planet. Valid planets are Swe.Constant.SE_SUN up to
  * Swe.Constant.SE_INTP_PERG with the exception of  Swe.Constant.SE_EARTH.
  * @param pl2 The second planet that will be transited by the first planet.
  * @param flags The calculation type flags (Swe.Constant.SEFLG_TRANSIT_LONGITUDE,
  * Swe.Constant.SEFLG_TRANSIT_LATITUDE or Swe.Constant.SEFLG_TRANSIT_DISTANCE in
  * conjunction with Swe.Constant.SEFLG_TRANSIT_SPEED for transits over a speed
  * value, and Swe.Constant.SEFLG_YOGA_TRANSIT to calculate for the SUM of the
  * two positions or speeds instead of the difference plus
  * Swe.Constant.SEFLG_PARTILE_TRANSIT, Swe.Constant.SEFLG_PARTILE_TRANSIT_START or
  * Swe.Constant.SEFLG_PARTILE_TRANSIT_END to calculate partile aspects). Also
  * flags modifying the basic planet calculations, these are
  * Swe.Constant.SEFLG_TOPOCTR, Swe.Constant.SEFLG_HELCTR, Swe.Constant.SEFLG_TRUEPOS and
  * Swe.Constant.SEFLG_SIDEREAL, plus the (optional)
//#ifdef NO_JPL
  * ephemeris flags Swe.Constant.SEFLG_MOSEPH, Swe.Constant.SEFLG_SWIEPH.
//#else
  * ephemeris flags Swe.Constant.SEFLG_MOSEPH, Swe.Constant.SEFLG_SWIEPH
  * or Swe.Constant.SEFLG_JPLEPH.<br>
  * SEFLG_PARTILE_TRANSIT will calculate the next time, when the planets will
  * have a partile aspect to each other, if they don't have a partile aspect
  * now. If they do have a partile aspect now, it calculates the next time,
  * when this partile aspect gets lost.
//#endif /* NO_JPL */
  //* @param offset This is an offset to the exact conjunction transit point.
  //* E.g., when the offset is 180 for longitude calculations, you will get the
  //* dates, when the two planets are opposite to each other. Note: The offset
  //* is related to the FIRST planet, so an offset value of 30 degree will find
  //* the transit points, when the FIRST planet will be 30 degrees after the
  //* position of the second planet.
  //* @see swisseph.TCPlanet#TCPlanet(SwissEph, int, int, double)
  //* @see swisseph.Swe.Constant#SEFLG_TRANSIT_LONGITUDE
  //* @see swisseph.Swe.Constant#SEFLG_TRANSIT_LATITUDE
  //* @see swisseph.Swe.Constant#SEFLG_TRANSIT_DISTANCE
  //* @see swisseph.Swe.Constant#SEFLG_TRANSIT_SPEED
  //* @see swisseph.Swe.Constant#SEFLG_PARTILE_TRANSIT
  //* @see swisseph.Swe.Constant#SEFLG_PARTILE_TRANSIT_START
  //* @see swisseph.Swe.Constant#SEFLG_PARTILE_TRANSIT_END
  //* @see swisseph.Swe.Constant#SEFLG_YOGA_TRANSIT
  //* @see swisseph.Swe.Constant#SEFLG_TOPOCTR
  //* @see swisseph.Swe.Constant#SEFLG_HELCTR
  //* @see swisseph.Swe.Constant#SEFLG_TRUEPOS
  //* @see swisseph.Swe.Constant#SEFLG_SIDEREAL
  //* @see swisseph.Swe.Constant#SEFLG_MOSEPH
  //* @see swisseph.Swe.Constant#SEFLG_SWIEPH
//#ifndef NO_JPL
//  * @see swisseph.Swe.Constant#SEFLG_JPLEPH
//#endif /* NO_JPL */
  public SwissEph sw;
  public TCPlanetPlanet(SwissEph sw,  int pl1, int pl2, int flags, double offset)
  {
    this.sw = sw;
    if (this.sw == null) {
      this.sw = new SwissEph();
    }

    // Check parameter: //////////////////////////////////////////////////////
    // List of all valid flags:
        
    this.tflags = flags;
        
    int vFlags = Transit.Constant.SEFLG_EPHMASK |
                 Swe.Constant.SEFLG_TOPOCTR |
                 Swe.Constant.SEFLG_HELCTR |
                 Swe.Constant.SEFLG_NOABERR |
                 Swe.Constant.SEFLG_NOGDEFL |
                 Swe.Constant.SEFLG_SIDEREAL |
                 Swe.Constant.SEFLG_TRUEPOS |
                 Transit.Constant.SEFLG_PARTILE_TRANSIT |
                 Transit.Constant.SEFLG_PARTILE_TRANSIT_START |
                 Transit.Constant.SEFLG_PARTILE_TRANSIT_END |
                 Transit.Constant.SEFLG_YOGA_TRANSIT |
                 Transit.Constant.SEFLG_TRANSIT_LONGITUDE |
                 Transit.Constant.SEFLG_TRANSIT_LATITUDE |
                 Transit.Constant.SEFLG_TRANSIT_DISTANCE |
                 Transit.Constant.SEFLG_TRANSIT_SPEED;
    // NOABERR and NOGDEFL is allowed for HELCTR, as they get set
    // anyway.
    if ((flags & Swe.Constant.SEFLG_HELCTR) != 0) {
      vFlags |= Swe.Constant.SEFLG_NOABERR | Swe.Constant.SEFLG_NOGDEFL;
    }
    //if ((flags&~vFlags) != 0) {
    //  throw new IllegalArgumentException("Invalid flag(s): "+(flags&~vFlags));
    //}

    // Allow only one of SEFLG_TRANSIT_LONGITUDE, SEFLG_TRANSIT_LATITUDE, SEFLG_TRANSIT_DISTANCE:
    int type = flags&(Transit.Constant.SEFLG_TRANSIT_LONGITUDE |
                      Transit.Constant.SEFLG_TRANSIT_LATITUDE |
                      Transit.Constant.SEFLG_TRANSIT_DISTANCE);
    //if (type != Swe.Constant.SEFLG_TRANSIT_LONGITUDE &&
    //    type != Swe.Constant.SEFLG_TRANSIT_LATITUDE &&
    //    type != Swe.Constant.SEFLG_TRANSIT_DISTANCE) {
    //  throw new IllegalArgumentException("Invalid flag combination '" + flags +
    //    "': specify exactly one of SEFLG_TRANSIT_LONGITUDE (" +
    //    Swe.Constant.SEFLG_TRANSIT_LONGITUDE + "), SEFLG_TRANSIT_LATITUDE (" +
    //    Swe.Constant.SEFLG_TRANSIT_LATITUDE + "), SEFLG_TRANSIT_DISTANCE (" +
    //    Swe.Constant.SEFLG_TRANSIT_DISTANCE + ").");
    //}
//    if (pl1<Swe.Constant.SE_SUN ||
////#ifdef ASTROLOGY
//        pl1>Swe.Constant.SE_MEAN_APOG) {
////#else
//        pl1>Swe.Constant.SE_INTP_PERG ||
//        pl1==Swe.Constant.SE_EARTH) {
////#endif /* ASTROLOGY */
//      throw new IllegalArgumentException(
//          "Unsupported planet number " + pl1 + " ("+
//              sw.swe_get_planet_name(pl1) + ")");
//    }
//    if (pl2<Swe.Constant.SE_SUN ||
////#ifdef ASTROLOGY
//        pl2>Swe.Constant.SE_MEAN_APOG) {
////#else
//        pl2>Swe.Constant.SE_INTP_PERG ||
//        pl2==Swe.Constant.SE_EARTH) {
////#endif /* ASTROLOGY */
//      throw new IllegalArgumentException(
//          "Unsupported planet number " + pl1 + " ("+
//              sw.swe_get_planet_name(pl1) + ")");
//    }
    //if ((flags & Swe.Constant.SEFLG_HELCTR) != 0 &&
    //    (pl1 == Swe.Constant.SE_MEAN_APOG ||
    //     pl1 == Swe.Constant.SE_OSCU_APOG ||
    //     pl1 == Swe.Constant.SE_MEAN_NODE ||
    //     pl1 == Swe.Constant.SE_TRUE_NODE)) {
    //  throw new IllegalArgumentException(
    //      "Unsupported planet number " + pl1 + " (" +
    //          sw.swe_get_planet_name(pl1) + ") for heliocentric " +
    //          "calculations");
    //}
    //if ((flags & Swe.Constant.SEFLG_HELCTR) != 0 &&
    //    (pl2 == Swe.Constant.SE_MEAN_APOG ||
    //     pl2 == Swe.Constant.SE_OSCU_APOG ||
    //     pl2 == Swe.Constant.SE_MEAN_NODE ||
    //     pl2 == Swe.Constant.SE_TRUE_NODE)) {
    //  throw new IllegalArgumentException(
    //      "Unsupported planet number " + pl2 + " (" +
    //          sw.swe_get_planet_name(pl2) + ") for heliocentric " +
    //          "calculations");
    //}
    //if (pl1 == pl2) {
    //  throw new IllegalArgumentException(
    //               "Transiting and referred planet have to be different!");
    //}

    this.pl1 = pl1;
    this.pl2 = pl2;

    calcPartile = ((flags & Transit.Constant.SEFLG_PARTILE_TRANSIT_START) != 0);	// Included in partileBoth
    calcNonPartile = ((flags & Transit.Constant.SEFLG_PARTILE_TRANSIT_END) != 0);	// Included in partileBoth
    calcPartileBoth = (calcPartile && calcNonPartile);
    calcYoga = ((flags & Transit.Constant.SEFLG_YOGA_TRANSIT) != 0);

    // The index into the xx[] array in swe_calc() to use:
    if ((flags&Transit.Constant.SEFLG_TRANSIT_LATITUDE) != 0) { // Calculate latitudinal transits
      idx = 1;
    } else if ((flags&Transit.Constant.SEFLG_TRANSIT_DISTANCE) != 0) { // Calculate distance transits
      idx = 2;
    }
    if ((flags&Transit.Constant.SEFLG_TRANSIT_SPEED) != 0) { // Calculate speed transits
      idx += 3;
      flags |= Swe.Constant.SEFLG_SPEED;
    }

    // Eliminate SEFLG_TRANSIT_* flags for use in swe_calc():
    flags &= ~(Transit.Constant.SEFLG_TRANSIT_LONGITUDE |
               Transit.Constant.SEFLG_TRANSIT_LATITUDE |
               Transit.Constant.SEFLG_TRANSIT_DISTANCE |
               Transit.Constant.SEFLG_PARTILE_TRANSIT |
               Transit.Constant.SEFLG_PARTILE_TRANSIT_START |
               Transit.Constant.SEFLG_PARTILE_TRANSIT_END |
               Transit.Constant.SEFLG_YOGA_TRANSIT |
               Transit.Constant.SEFLG_TRANSIT_SPEED);
    this.flags = flags;

    // Calculate basic parameters: ///////////////////////////////////////////
    rollover = (idx == 0);

    //if (calcPartile || calcNonPartile) {
    //  if ((offset % 30) != 0) {
    //    throw new IllegalArgumentException(
    //                 "Wrong offset (" + offset + "). Calculation of partile aspect may have offsets of multiples of 30 degrees only.");
    //  }
    //  rollover = false;
    //}
    this.offset = checkOffset(offset);


    maxSpeed1=getSpeed(false,pl1);
    minSpeed1=getSpeed(true,pl1);
    maxSpeed2=getSpeed(false,pl2);
    minSpeed2=getSpeed(true,pl2);
    //if (Double.isInfinite(maxSpeed1) || Double.isInfinite(minSpeed1)) {
    //  throw new IllegalArgumentException(
    //      ((flags&Swe.Constant.SEFLG_TOPOCTR)!=0?"Topo":((flags&Swe.Constant.SEFLG_HELCTR)!=0?"Helio":"Geo")) +
    //      "centric transit calculations with planet number " + pl1 + " ("+
    //      sw.swe_get_planet_name(pl1) + ") not possible: extreme speeds of the planet not available.");
    //}
    //if (Double.isInfinite(maxSpeed2) || Double.isInfinite(minSpeed2)) {
    //  throw new IllegalArgumentException(
    //      ((flags&Swe.Constant.SEFLG_TOPOCTR)!=0?"Topo":((flags&Swe.Constant.SEFLG_HELCTR)!=0?"Helio":"Geo")) +
    //      "centric transit calculations with planet number " + pl2 + " ("+
    //      sw.swe_get_planet_name(pl2) + ") not possible: extreme speeds of the planet not available.");
    //}

    if (calcYoga) {
      minSpeed = minSpeed1+minSpeed2;
      maxSpeed = maxSpeed1+maxSpeed2;
    } else {
      if (rollover) {
        minSpeed = (maxSpeed1>maxSpeed2)?minSpeed1-maxSpeed2:minSpeed2-maxSpeed1;
        maxSpeed = (maxSpeed1>maxSpeed2)?maxSpeed1-minSpeed2:maxSpeed2-minSpeed1;
      } else {
        minSpeed = Math.Min(minSpeed1-maxSpeed2, minSpeed2-maxSpeed1);
        maxSpeed = Math.Max(maxSpeed1-minSpeed2, maxSpeed2-minSpeed1);
      }
    }
  }

  // Looks for next time when the condition (partile / non-partile) is NOT met,
  // so we have a starting point for partile / non-partile calculation:
  public override double preprocessDate(double jdET, bool back)
  {
    if (calcPartile || calcNonPartile) {
      // Check the current partile status:
      isPartile = hasPartileAspect(jdET, pl1, pl2, flags, offset);

      if (!calcPartileBoth) {
        if (isPartile && calcPartile) {
          // Set next non partile date as starting date
          int flgs = flags & ~Transit.Constant.SEFLG_PARTILE_TRANSIT_START;
          flgs |= Transit.Constant.SEFLG_PARTILE_TRANSIT_END | Transit.Constant.SEFLG_TRANSIT_LONGITUDE;
          TransitCalculator tc = new TCPlanetPlanet(
              sw,
              pl1,
              pl2,
              flgs,
              offset);

          jdET = sw.getTransitET(tc, jdET, back);
          isPartile = !isPartile;
        } else if (!isPartile && calcNonPartile) {
          // Set next partile date as starting date
          int flgs = flags & ~Transit.Constant.SEFLG_PARTILE_TRANSIT_END;
          flgs |= Transit.Constant.SEFLG_PARTILE_TRANSIT_START | Transit.Constant.SEFLG_TRANSIT_LONGITUDE;
          TransitCalculator tc = new TCPlanetPlanet(
              sw,
              pl1,
              pl2,
              flgs,
              offset);

          jdET = sw.getTransitET(tc, jdET, back);
          isPartile = !isPartile;
        }
      }
    }
    return jdET;
  }

  /**
  * @return Returns true, if one position value is identical to another
  * position value. E.g., 360 degree is identical to 0 degree in
  * circle angles.
  * @see #rolloverVal
  */
  public override bool getRollover() {
    return rollover;
  }
  /**
  * This sets the transit degree or other transit value for the difference
  * or sum of the positions or speeds of both planets. It will be used on
  * the next call to getTransit().
  * @param value The offset value.
  * @see #getOffset()
  */
  public override void setOffset(double value)
  {
    offset = checkOffset(value);
  }
  /**
  * This returns the transit degree or other transit value of the relative
  * position or speed of the two planets.
  * @return The current offset value.
  * @see #setOffset(double)
  */
  public override double getOffset()
  {
    return offset;
  }
  /**
  * This returns the two planet numbers involved as Strings.
  * @return An array of identifiers identifying the calculated objects.
  */
  public override int[] getObjectIdentifiers()
  {
    return new int[]{ pl1,  pl2};
  }
//#ifdef EXTPRECISION
  /**
  * Set the factor, when to stop the calculation.
  * The getTransit*() methods will iterate calculations until the maximum
  * precision in the planetary calculation routines has been passed. With
  * this method, you specify a factor to the calculation precision used.
  * E.g., 100 means, stop the calculation, when the difference between the
  * calculated value and requested value has crossed a value of 100 times
  * less than the precision available in the calculation routines. 0.01 on
  * the other hand would stop the calculation, BEFORE the maximum available
  * precision had been reached. The default for the precision factor is 1.<p>
  * @param pfac The factor for the precision as explained above.
  * <b>Note:</b> A value greater than one will NOT really increase precision,
  * it will just <i>appear</i> to do so. It can be handy to show more equal
  * values when approaching the transit point from different sides (forward
  * or backward) or from different starting points.
  * @see #getPrecisionFactor()
  */
  public override void setPrecisionFactor(double pfac)
  {
    extPrecision = pfac;
  }
  /**
  * Returns the factor used to control stopping of the calculation
  * iterations.
  * @return The precision factor as set by the setPrecisionFactor()
  * method.
  * @see #setPrecisionFactor(double)
  */
  public override double getPrecisionFactor()
  {
    return extPrecision;
  }
//#endif /* EXTPRECISION */

  /**
  * Checks if the two planets have a partile aspect.
  */
  public bool hasPartileAspect(double jdET, int p1, int p2, int flgs, double offset) {
    StringBuilder serr = new StringBuilder();
    double[] xx1 = new double[6], xx2 = new double[6];
    offset = (int)offset;

    flgs &= ~(Transit.Constant.SEFLG_TRANSIT_LONGITUDE |
               Transit.Constant.SEFLG_TRANSIT_LATITUDE |
               Transit.Constant.SEFLG_TRANSIT_DISTANCE |
               Transit.Constant.SEFLG_PARTILE_TRANSIT |
               Transit.Constant.SEFLG_PARTILE_TRANSIT_START |
               Transit.Constant.SEFLG_PARTILE_TRANSIT_END |
               Transit.Constant.SEFLG_YOGA_TRANSIT |
               Transit.Constant.SEFLG_TRANSIT_SPEED);

    int ret = Swe.Function.swe_calc(jdET, p1, flgs, xx1, serr);
    //if (ret<0) {
    //  throw new SwissephException(jdET, SwissephException.UNDEFINED_ERROR,
    //      "Calculation failed with return code " + ret + ":\n" +
    //      serr.toString());
    //}

    ret = Swe.Function.swe_calc(jdET, p2, flgs, xx2, serr);
    //if (ret<0) {
    //  throw new SwissephException(jdET, SwissephException.UNDEFINED_ERROR,
    //      "Calculation failed with return code " + ret + ":\n" +
    //      serr.toString());
    //}

      return (int)(xx1[0]%30) == (int)(xx2[0]%30) &&
          ((((int)(xx1[0]) + offset + rolloverVal) % rolloverVal == (int)xx2[0]) ||
           (((int)(xx1[0]) - offset + rolloverVal) % rolloverVal == (int)xx2[0]));
  }




  ///////////////////////////////////////////////////////////////////////////////
  public override double getMaxSpeed()
  {
    return maxSpeed;
  }
  public override double getMinSpeed()
  {
    return minSpeed;
  }

  public override double calc(double jdET)
  {
    StringBuilder serr = new StringBuilder();
    double[] xx1 = new double[6];
    double[] xx2 = new double[6];

    int ret = Swe.Function.swe_calc(jdET, pl1, flags, xx1, serr);
    //if (ret<0) {
    //  throw new SwissephException(jdET, SwissephException.UNDEFINED_ERROR,
    //      "Calculation failed with return code " + ret + ":\n" +
    //      serr.toString());
    //}

    ret = Swe.Function.swe_calc(jdET, pl2, flags, xx2, serr);
    //if (ret<0) {
    //  throw new SwissephException(jdET, SwissephException.UNDEFINED_ERROR,
    //      "Calculation failed with return code " + ret + ":\n" +
    //      serr.toString());
    //}

    if (calcPartile || calcNonPartile) {
      lon1 = xx1[0];
      lon2 = xx2[0];
      double delta1, delta2, delta3, delta4;
      double target1 = ((int)(lon1) + offset + rolloverVal) % rolloverVal;
      double target2 = ((int)(lon1) - offset + rolloverVal) % rolloverVal;
      double target3 = ((int)(lon2) + offset + rolloverVal) % rolloverVal;
      double target4 = ((int)(lon2) - offset + rolloverVal) % rolloverVal;
      if (lon2 > target1) {
        delta1 = Math.Abs(lon2 - target1 - 1 + 1E-9);	// Might be better to use a different comparison operator instead of  '+ 1E-9'...?
      } else {
        delta1 = target1 - lon2;
      }
      if (lon2 > target2) {
        delta2 = Math.Abs(lon2 - target2 - 1 + 1E-9);
      } else {
        delta2 = target2 - lon2;
      }
      if (lon1 > target3) {
        delta3 = Math.Abs(lon1 - target3 - 1 + 1E-9);
      } else {
        delta3 = target3 - lon1;
      }
      if (lon1 > target4) {
        delta4 = Math.Abs(lon1 - target4 - 1 + 1E-9);
      } else {
        delta4 = target4 - lon1;
      }

      double diff = Math.Min(Math.Min(Math.Min(delta1, delta2), delta3), delta4) % 1;	// Using '%1', as I don't get the next degree value, where these planets will be partile (on calcPartile?).
      if ((!isPartile && calcPartile && diff == 0) || (isPartile && calcNonPartile && diff != 0)) {
        return offset;
      }
      if (((int)xx1[idx] - (int)xx2[idx] + rolloverVal) %rolloverVal - offset == 0 ||
          ((int)xx2[idx] - (int)xx1[idx] + rolloverVal) %rolloverVal - offset == 0) {
        return Double.PositiveInfinity;	// Means: found, but don't interpolate with previous values
      }
      return ((xx1[idx] % rolloverVal) - (xx2[idx] % rolloverVal) + rolloverVal) % rolloverVal;
    } else if (calcYoga) {
//System.err.println(xx1[idx] + " + " + xx2[idx] + " = " + (xx1[idx] + xx2[idx]));
//        return xx1[idx] + xx2[idx];
    }
    return xx1[idx] - xx2[idx];
  }



  public override double getTimePrecision(double degPrec)
  {
    // Recalculate degPrec to mean the minimum  time, in which the planet can
    // possibly move that degree:
    double amin = Math.Min(Math.Abs(minSpeed1),Math.Abs(minSpeed2));
    double amax = Math.Min(Math.Abs(maxSpeed1),Math.Abs(maxSpeed2));

    double maxVal = Math.Max(Math.Abs(amin),Math.Abs(amax));
    if (maxVal != 0.0) {
      return degPrec / maxVal;
    }
    return 1E-9;
  }

  public override double getDegreePrecision(double jd)
  {
    // Calculate the planet's minimum movement regarding the maximum available
    // precision.
    //
    // For all calculations, we assume the following minimum exactnesses
    // based on the discussions on http://www.astro.com/swisseph, even though
    // these values are nothing more than very crude estimations which should
    // leave us on the save side always, even more, when seeing that we always
    // consider the maximum possible speed / acceleration of a planet in the
    // transit calculations and not the real speed.
    //
    // Take degPrec to be the minimum exact degree in longitude
    double degPrec = 0.005;
    if (idx>2) { // Speed
      // "The speed precision is now better than 0.002" for all planets"
      degPrec = 0.002;
    } else { // Degrees
      // years 1980 to 2099:              0.005"
      // years before 1980:               0.08"   (from sun to jupiter)
      // years 1900 to 1980:              0.08"   (from saturn to neptune) (added: nodes)
      // years before 1900:               1"      (from saturn to neptune) (added: nodes)
      // years after 2099:                same as before 1900
      //
      if (pl1>=Swe.Constant.SE_SUN && pl1<=Swe.Constant.SE_JUPITER) {
        if (jd<1980 || jd>2099) {
          degPrec = 0.08;
        }
      } else {
        if (jd>=1900 && jd<1980) {
          degPrec = 0.08;
        } else if (jd<1900 || jd>2099) { // Unclear about true nodes...
          degPrec = 1;
        }
      }
      if (pl2>=Swe.Constant.SE_SUN && pl2<=Swe.Constant.SE_JUPITER) {
        if (jd<1980 || jd>2099) {
          degPrec = Math.Max(0.08,degPrec);
        }
      } else {
        if (jd>=1900 && jd<1980) {
          degPrec = Math.Max(0.08,degPrec);
        } else if (jd<1900 || jd>2099) { // Unclear about true nodes...
          degPrec = Math.Max(1,degPrec);
        }
      }
    }
    degPrec/=3600.0;
    degPrec*=0.5; // We take the precision to be BETTER THAN ... as it is stated somewhere

    // We recalculate these degrees to the minimum time difference that CAN
    // possibly give us data differing more than the above given precision.
    switch (idx) {
      case 0: // Longitude
            break;
      case 1: // Latitude
            break;
      case 3: // Speed in longitude
            break;
      case 4: // Speed in latitude
            break;
      case 2: // Distance
            break;
      case 5: // Speed in distance
        // We need to recalculate the precision in degrees to a distance value.
        // For this we need the maximum distance to the centre of calculation,
        // which is the barycentre for the main planets.
        degPrec *= Math.Max(sw.ext.maxBaryDist[pl1],
                            sw.ext.maxBaryDist[pl2]);
        break;
    }

    return degPrec;

    // Barycentre:
    //            0.981683040      1.017099581  (Barycenter of the earth!)
    // Sun:       0.982747149 AU   1.017261973 AU
    // Moon:      0.980136691 AU   1.019846623 AU
    // Mercury:   0.307590579 AU   0.466604085 AU
    // Venus:     0.717960758 AU   0.728698831 AU
    // Mars:      1.382830768 AU   0.728698831 AU
    // Jupiter:   5.448547595 AU   4.955912195 AU
    // Saturn:   10.117683425 AU   8.968685733 AU
    // Uranus:   18.327870391 AU  19.893326756 AU
    // Neptune:  29.935653168 AU  30.326750627 AU
    // Pluto:    29.830132096 AU  41.499626899 AU
    // MeanNode:  0.002569555 AU   0.002569555 AU
    // TrueNode:  0.002361814 AU   0.002774851 AU

    //
    // Minimum and maximum (barycentric) distances:
    // Sun:       0.000095 AU      0.01034 AU
    // Moon:      0.972939 AU      1.02625 AU
    // Mercury:   0.298782 AU      0.47569 AU
    // Venus:     0.709190 AU      0.73723 AU
    // Mars:      1.370003 AU      1.67685 AU
    // Jupiter:   4.912031 AU      5.47705 AU
    // Saturn:    8.948669 AU     10.13792 AU
    // Uranus:   18.257511 AU     20.12033 AU
    // Neptune:  29.780622 AU     30.36938 AU
    // Pluto:    29.636944 AU     49.43648 AU
    // MeanNode:  -        AU      -       AU ?
    // TrueNode:  -        AU      -       AU ?


    // Maximum and minimum (geocentric) distances:
    // Sun:        1.016688129 AU   0.983320477 AU
    // Moon:       0.002710279 AU   0.002439921 AU
    // Mercury:    0.549188094 AU   1.448731236 AU

    // Saturn:     7.84 / 7.85 AU  11.25/11.26  AU
    // Uranus:    21.147/21.148 AU              AU

  }


  //////////////////////////////////////////////////////////////////////////////

  private double checkOffset(double val) {
    // Similar rollover considerations for the latitude will be necessary, if
    // swe_calc() would return latitudinal values beyond -90 and +90 degrees.

    if (rollover) {        // Longitude from 0 to 360 degrees:
      while (val < 0.0) { val += rolloverVal; }
      val %= rolloverVal;
      minVal = 0.0;
      maxVal = rolloverVal;
    } else if (idx == 1) { // Latitude from -90 to +90 degrees:
      while (val < -90.0) { val += 180.0; }
      while (val >  90.0) { val -= 180.0; }
      minVal = -90.0;
      maxVal = +90.0;
    }
    return val;
  }


  private double getSpeed(bool min, int planet) {
    bool lat = ((tflags&Transit.Constant.SEFLG_TRANSIT_LATITUDE) != 0);
    bool dist = ((tflags&Transit.Constant.SEFLG_TRANSIT_DISTANCE) != 0);
    bool speed = ((tflags&Transit.Constant.SEFLG_TRANSIT_SPEED) != 0);
    bool topo = ((tflags&Swe.Constant.SEFLG_TOPOCTR) != 0);
    bool helio = ((tflags&Swe.Constant.SEFLG_HELCTR) != 0);

    // Some topocentric speeds are very different to the geocentric
    // speeds, so we use other values than for geocentric calculations:
    if (topo) {
    //  if (!sw.swed.geopos_is_set) {
    //    throw new IllegalArgumentException("Geographic position is not set for "+
    //                                       "requested topocentric calculations.");
    //  }
    //  if (sw.swed.topd.geoalt>50000.) {
    //    throw new IllegalArgumentException("Topocentric transit calculations "+
    //                                       "are restricted to a maximum "+
    //                                       "altitude of 50km so far.");
    //  } else if (sw.swed.topd.geoalt<-12000000) {
    //    throw new IllegalArgumentException("Topocentric transit calculations "+
    //                                       "are restricted to a minimum "+
    //                                       "altitude of -12000km so far.");
    //  }
      
      if (speed) {
        if (lat) {
          return (min?Transit.Constant.minTopoLatAccel[planet]:Transit.Constant.maxTopoLatAccel[planet]);
        } else if (dist) {
          return (min?Transit.Constant.minTopoDistAccel[planet]:Transit.Constant.maxTopoDistAccel[planet]);
        } else {
          return (min?Transit.Constant.minTopoLonAccel[planet]:Transit.Constant.maxTopoLonAccel[planet]);
        }
      } else {
        if (lat) {
          return (min?Transit.Constant.minTopoLatSpeed[planet]:Transit.Constant.maxTopoLatSpeed[planet]);
        } else if (dist) {
          return (min?Transit.Constant.minTopoDistSpeed[planet]:Transit.Constant.maxTopoDistSpeed[planet]);
        } else {
          return (min?Transit.Constant.minTopoLonSpeed[planet]:Transit.Constant.maxTopoLonSpeed[planet]);
        }
      }
    }
    
    // Heliocentric speeds are very different to the geocentric speeds, so
    // we use other values than for geocentric calculations:
    if (helio) {
      if (speed) {
        if (lat) {
          return (min?Transit.Constant.minHelioLatAccel[planet]:Transit.Constant.maxHelioLatAccel[planet]);
        } else if (dist) {
          return (min?Transit.Constant.minHelioDistAccel[planet]:Transit.Constant.maxHelioDistAccel[planet]);
        } else {
          return (min?Transit.Constant.minHelioLonAccel[planet]:Transit.Constant.maxHelioLonAccel[planet]);
        }
      } else {
        if (lat) {
          return (min?Transit.Constant.minHelioLatSpeed[planet]:Transit.Constant.maxHelioLatSpeed[planet]);
        } else if (dist) {
          return (min?Transit.Constant.minHelioDistSpeed[planet]:Transit.Constant.maxHelioDistSpeed[planet]);
        } else {
          return (min?Transit.Constant.minHelioLonSpeed[planet]:Transit.Constant.maxHelioLonSpeed[planet]);
        }
      }
    }

    // Geocentric:
    if (speed) {
      if (lat) {
          return (min ? Transit.Constant.minLatAccel[planet] : Transit.Constant.maxLatAccel[planet]);
      } else if (dist) {
          return (min ? Transit.Constant.minDistAccel[planet] : Transit.Constant.maxDistAccel[planet]);
      } else {
          return (min ? Transit.Constant.minLonAccel[planet] : Transit.Constant.maxLonAccel[planet]);
      }
    } else {
      if (lat) {
          return (min ? Transit.Constant.minLatSpeed[planet] : Transit.Constant.maxLatSpeed[planet]);
      } else if (dist) {
          return (min ? Transit.Constant.minDistSpeed[planet] : Transit.Constant.maxDistSpeed[planet]);
      } else {
          return (min ? Transit.Constant.minLonSpeed[planet] : Transit.Constant.maxLonSpeed[planet]);
      }
    }
  }

  protected bool checkIdenticalResult(double offset, double val) {
    if (calcPartile) {
      bool res = (((lon1 % 1) < 1E-9 || (lon2 % 1) < 1E-9) &&	// leave other values to the interpolation after the checkResult() call otherwise
            (int)(lon1%30) == (int)(lon2%30) &&
          ((((int)(lon1) + offset + rolloverVal) % rolloverVal == (int)lon2) ||
           (((int)(lon1) - offset + rolloverVal) % rolloverVal == (int)lon2)));
      return res;
    } else if (calcNonPartile) {
      bool res = (((lon1 % 1) < 1E-9 && (lon2 % 1) < 1E-9) ||	// leave other values to the interpolation after the checkResult() call otherwise
           (int)(lon1%30) != (int)(lon2%30) ||
          ((((int)(lon1) + offset + rolloverVal) % rolloverVal != (int)lon2) &&
           (((int)(lon1) - offset + rolloverVal) % rolloverVal != (int)lon2)));
      return res;
    }
    return val == offset;
  }

  protected bool checkResult(double offset, double lastVal, double val, bool above, bool pxway) {
    if (calcPartile) {
      bool res = ((int)(lon1%30) == (int)(lon2%30) &&
          ((((int)(lon1) + offset + rolloverVal) % rolloverVal == (int)lon2) ||
           (((int)(lon1) - offset + rolloverVal) % rolloverVal == (int)lon2)));
      return res;
    } else if (calcNonPartile) {
      bool res = ((int)(lon1%30) != (int)(lon2%30) ||
          ((((int)(lon1) + offset + rolloverVal) % rolloverVal != (int)lon2) &&
           (((int)(lon1) - offset + rolloverVal) % rolloverVal != (int)lon2)));
      return res;
    }
    return checkResult(offset, lastVal, val, above, pxway);
  }
  protected double getNextJD(double jdET, double val, double offset, double min, double max, bool back) {
    if (calcPartile || calcNonPartile) {
      double delta1, delta2, delta3, delta4;
      double deltaET;

      // direct and direct motion:
      if (calcPartile) {
        // distances in degree to next possible matching (int-)degree
        double target1 = ((int)(lon1) + offset + rolloverVal) % rolloverVal;
        double target2 = ((int)(lon1) - offset + rolloverVal) % rolloverVal;
        double target3 = ((int)(lon2) + offset + rolloverVal) % rolloverVal;
        double target4 = ((int)(lon2) - offset + rolloverVal) % rolloverVal;

        if (lon2 > target1) {
          delta1 = Math.Abs(lon2 - target1 - 1 + 1E-9);
        } else {
          delta1 = target1 - lon2;
        }
        if (lon2 > target2) {
          delta2 = Math.Abs(lon2 - target2 - 1 + 1E-9);
        } else {
          delta2 = target2 - lon2;
        }
        if (lon1 > target3) {
          delta3 = Math.Abs(lon1 - target3 - 1 + 1E-9);
        } else {
          delta3 = target3 - lon1;
        }
        if (lon1 > target4) {
          delta4 = Math.Abs(lon1 - target4 - 1 + 1E-9);
        } else {
          delta4 = target4 - lon1;
        }
      } else {	// partile end
        // distances in degree to next possible non-matching degrees
        // ----|->---|-----------------------------------------------|--->-|----------------------
        delta1 = lon1 - (int)lon1;
        delta2 = 1 - delta1;
        delta3 = lon2 - (int)lon2;
        delta4 = 1 - delta3;
        delta1 += 1E-9;
        delta2 += 1E-9;
        delta3 += 1E-9;
        delta4 += 1E-9;
      }

      double maxSpeed = Math.Abs(maxSpeed2 - minSpeed1);
      maxSpeed = Math.Max(maxSpeed, Math.Abs(maxSpeed1 - minSpeed2));
      double minDx = Math.Min(Math.Min(Math.Min(delta1, delta2), delta3), delta4)%1;
      deltaET = Math.Abs(minDx / maxSpeed);

      return jdET + (back ? - deltaET : deltaET);
    }

    return getNextJD(jdET, val, offset, min, max, back);
  }


  //public String toString() {
  //  return "[Planets:" + pl1 + "/" + pl2 + "];Offset:" + getOffset();
  //}

    }
}
